import hydra
import jax
import omegaconf

from compass.trainers.validation import validate
from compass.trainers.validation_jobshop import validate as validate_jssp
from compass.utils.logger import EnsembleLogger, NeptuneLogger, TerminalLogger


def create_logger(cfg) -> EnsembleLogger:
    loggers = []

    if "terminal" in cfg.logger:
        loggers.append(TerminalLogger(**cfg.logger.terminal))

    if "neptune" in cfg.logger:
        neptune_config = {}
        neptune_config["name"] = cfg.logger.neptune.name
        neptune_config["project"] = cfg.logger.neptune.project
        neptune_config["tags"] = [f"{cfg.algo_name}", "fastrl", f"{cfg.env_name}"]
        neptune_config["parameters"] = cfg

        loggers.append(NeptuneLogger(**neptune_config))

    # return the loggers
    return EnsembleLogger(loggers)


@hydra.main(
    config_path="config",
    version_base=None,
    config_name="config_exp",
)
def run(cfg: omegaconf.DictConfig) -> None:

    # create the name of the run's directory - used for logging and checkpoints
    run_subdirectory = (
        str(cfg.env_name)
        + "/"
        + str(cfg.algo_name)
        + "/"
        + f"bs{cfg.batch_size}_tss{cfg.training_sample_size}"
        + f"_ga{cfg.optimizer.num_gradient_accumulation_steps}"
        + f"_bd{cfg.behavior_dim}"
        + f"_ba{cfg.behavior_amplification}_seed{cfg.seed}/"
    )

    cfg.checkpointing.directory = cfg.checkpointing.directory + run_subdirectory
    cfg.validation.checkpointing.restore_path = (
        cfg.validation.checkpointing.restore_path + run_subdirectory
    )

    # Check and configure the available devices.
    behavior_dim = cfg.behavior_dim
    validation_cfg = cfg.validation

    available_devices = len(jax.local_devices())
    if validation_cfg.num_devices < 0:
        validation_cfg.num_devices = available_devices
        print(f"Using {available_devices} available device(s).")
    else:
        assert (
            available_devices >= validation_cfg.num_devices
        ), f"{validation_cfg.num_devices} devices requested but only {available_devices} available."

    # create a logger
    cfg.logger.neptune.name = "fastrl-" + cfg.logger.neptune.name
    logger = create_logger(cfg)

    # init the random key
    random_key = jax.random.PRNGKey(validation_cfg.problem_seed)

    if cfg.env_name == "jssp":
        metrics = validate_jssp(
            random_key=random_key,
            cfg=validation_cfg,
            behavior_dim=behavior_dim,
        )
    else:
        metrics = validate(
            random_key=random_key,
            cfg=validation_cfg,
            behavior_dim=behavior_dim,
        )

    metrics = {f"validate/{k}": v for (k, v) in metrics.items()}
    logger.write(metrics)


if __name__ == "__main__":
    run()
